`default_nettype none

module MAP04(   //signal descriptions in powerpak.v
    input m2,
    input m2_n,
    input clk20,

    input reset,
    input nesprg_we,
    output nesprg_oe,
    input neschr_rd,
    input neschr_wr,
    input [15:0] prgain,
    input [13:0] chrain,
    input [7:0] nesprgdin,
    input [7:0] ramprgdin,
    output [7:0] nesprgdout,

    output [7:0] neschrdout,
    output neschr_oe,

    output chrram_we,
    output chrram_oe,
    output wram_oe,
    output wram_we,
    output prgram_we,
    output prgram_oe,
    output [18:10] ramchraout,
    output [18:13] ramprgaout,
    output reg irq,
    output ciram_ce,
    output exp6,
    
    input cfg_boot,
    input [18:12] cfg_chrmask,
    input [18:13] cfg_prgmask,
    input cfg_vertical,
    input cfg_fourscreen,
    input cfg_chrram
);

    reg prgswap, chrswap, mirror;
    reg [7:0] irqlatch;
    reg [1:0] wramen;
    reg [2:0] regsel;
    reg [7:0] bankreg[7:0];
    reg irqen;
    always@(posedge m2) begin
        if(nesprg_we & prgain[15]) begin
            case({prgain[14:13],prgain[0]})
                0:{chrswap, prgswap, regsel} <= {nesprgdin[7:6],nesprgdin[2:0]};    //8000
                1:bankreg[regsel]<=nesprgdin;                                       //8001
                2:mirror<=nesprgdin[0] & ~cfg_fourscreen;                           //A000
                3:wramen<=nesprgdin[7:6];                                           //A001
                4:irqlatch<=nesprgdin;                                              //C000
                6:irqen<=0;
                7:irqen<=1;
            endcase
        end
    end

//irq

    reg a12;
    wire a12clear;
    always@(posedge chrain[12], posedge a12clear)
        if(a12clear)
            a12<=0;
        else
            a12<=1;

    reg [2:0] a12count;
    always@(posedge m2)
        if(chrain[12])
            a12count<=0;
        else if(a12count!=7)
            a12count<=a12count+1;
    assign a12clear=a12count==6;

    wire rset=nesprg_we & m2_n & {prgain[15:13],prgain[0]}==4'b1101;
    reg reload;
    always@(posedge a12, posedge rset)
        if(rset) reload<=1;
        else reload<=0;

    reg [7:0] count;
    always@(posedge a12) begin
        if(!count | reload)
            count<=irqlatch;
        else
            count<=count-1;
    end
    
    wire irqtrip=~|count;
    wire irqclear=nesprg_we & m2_n & {prgain[15:13],prgain[0]}==4'b1110;
    always@(posedge irqtrip, posedge irqclear)
        if(irqclear) irq<=0;
        else irq<=irqen;

//bankswitch
    reg [18:13] prgbank;
    always@(prgswap,prgain,bankreg[6],bankreg[7])
        case({prgswap,prgain[14:13]})
            0,6:prgbank=bankreg[6];
            1,5:prgbank=bankreg[7];
            2,4:prgbank=6'b111110;
            3,7:prgbank=6'b111111;
        endcase

    reg [18:10] chrbank;
    always@(cfg_fourscreen,mirror,chrswap,chrain,bankreg[0],bankreg[1],bankreg[2],bankreg[3],bankreg[4],bankreg[5]) begin
        chrbank[18]=chrain[13];
        if(!chrain[13]) begin
            case({chrswap,chrain[12:10]})
                0,1,12,13:chrbank[17:10]={bankreg[0][7:1],chrain[10]};
                2,3,14,15:chrbank[17:10]={bankreg[1][7:1],chrain[10]};
                4,8:chrbank[17:10]=bankreg[2];
                5,9:chrbank[17:10]=bankreg[3];
                6,10:chrbank[17:10]=bankreg[4];
                7,11:chrbank[17:10]=bankreg[5];
            endcase
        end else begin
            chrbank[17:12]=0;
            chrbank[11]=chrain[11] & cfg_fourscreen;
            chrbank[10]=((mirror & chrain[11]) | (~mirror & chrain[10]));
        end
    end

//rom size mask
    assign ramprgaout[18:13]=(prgbank[18:13] & cfg_prgmask & {4'b1111, {2{prgain[15]}}});
    assign ramchraout=chrbank & {1'b1, cfg_chrmask[17:12], 2'b11};

//bus control
    assign ciram_ce=0;
    assign chrram_we=neschr_wr & (cfg_chrram | chrain[13]);
    assign chrram_oe=neschr_rd;

    assign prgram_oe= ~cfg_boot & m2_n & ~nesprg_we & prgain[15];
    assign wram_oe=               m2_n & ~nesprg_we & prgain[15:13]==3'b011 /* & wramen[1] */;

    assign wram_we=m2_n & nesprg_we & prgain[15:13]=='b011 /* & wramen==2 */;   //wram protection breaks startropics / startropics2

    assign prgram_we=0;

    wire config_rd;
    gamegenie gg(m2, reset, nesprg_we, prgain, nesprgdin, ramprgdin, nesprgdout, config_rd);

    assign nesprg_oe=wram_oe | prgram_oe | config_rd;

    assign neschr_oe=0;
    assign neschrdout=0;

//sound
    assign exp6=1'bz;

endmodule