`default_nettype none
//Sunsoft 5B "Gimmick"

module MAP45(       //signal descriptions in powerpak.v
    input m2,
    input m2_n,
    input clk20,

    input reset,
    input nesprg_we,
    output nesprg_oe,
    input neschr_rd,
    input neschr_wr,
    input [15:0] prgain,
    input [13:0] chrain,
    input [7:0] nesprgdin,
    input [7:0] ramprgdin,
    output [7:0] nesprgdout,

    output [7:0] neschrdout,
    output neschr_oe,

    output chrram_we,
    output chrram_oe,
    output wram_oe,
    output wram_we,
    output prgram_we,
    output prgram_oe,
    output [18:10] ramchraout,
    output [18:13] ramprgaout,
    output irq,
    output ciram_ce,
    output exp6,

    input cfg_boot,
    input [18:12] cfg_chrmask,
    input [18:13] cfg_prgmask,
    input cfg_vertical,
    input cfg_fourscreen,
    input cfg_chrram
);
    reg [1:0] sramEN;
    reg [3:0] regA;
    reg [7:0] chrbank0, chrbank1, chrbank2, chrbank3, chrbank4, chrbank5, chrbank6, chrbank7;
    reg [5:0] prgbank6, prgbank8, prgbankA, prgbankC;
    reg [1:0] mirror;
    always@(posedge m2) begin
        if(nesprg_we) begin
            if(prgain[15:13]==3'b100)   //8000
                regA<=nesprgdin;
            if(prgain[15:13]==3'b101)   //A000
                case(regA)
                    0:chrbank0<=nesprgdin;
                    1:chrbank1<=nesprgdin;
                    2:chrbank2<=nesprgdin;
                    3:chrbank3<=nesprgdin;
                    4:chrbank4<=nesprgdin;
                    5:chrbank5<=nesprgdin;
                    6:chrbank6<=nesprgdin;
                    7:chrbank7<=nesprgdin;
                    8:{sramEN,prgbank6}<=nesprgdin;
                    9:prgbank8<=nesprgdin;
                    10:prgbankA<=nesprgdin;
                    11:prgbankC<=nesprgdin;
                    12:mirror<=nesprgdin;
                    //13:{timerEN,timerIRQEN}<={nesprgdin[7],nesprgdin[0]};
                    //14:timer[7:0]<=nesprgdin;
                    //15:timer[15:8]<=nesprgdin;
                endcase
        end
    end

//bankswitch
    reg [18:13] prgbank;
    reg [18:10] chrbank;
    always@* begin
        chrbank[18]=0;
        case(prgain[14:13])
            0:prgbank=prgbank8;
            1:prgbank=prgbankA;
            2:prgbank=prgbankC;
            3:prgbank={6{prgain[15]}}|prgbank6;
        endcase
        case(chrain[12:10])
            0:chrbank=chrbank0;
            1:chrbank=chrbank1;
            2:chrbank=chrbank2;
            3:chrbank=chrbank3;
            4:chrbank=chrbank4;
            5:chrbank=chrbank5;
            6:chrbank=chrbank6;
            7:chrbank=chrbank7;
        endcase
    end

//IRQ
    reg [15:0] timer;
    wire [16:0] timer_dec = timer-1;
    reg tripped;
    reg timerEN,timerIRQEN;
    always@(posedge m2, posedge reset) begin
        if(reset) begin
            timerIRQEN<=0;
            timerEN<=0;
            timer<=0;
            tripped<=0;
        end else begin
            if(nesprg_we & prgain[15:13]==3'b101 & regA==13) //write A000:D
                {timerEN,timerIRQEN}<={nesprgdin[7],nesprgdin[0]};

            if(nesprg_we & prgain[15:13]==3'b101 & regA==14) //write A000:E
                timer[7:0]<=nesprgdin;
            else if(timerEN)
                timer[7:0]<=timer_dec[7:0];

            if(nesprg_we & prgain[15:13]==3'b101 & regA==15) //write A000:F
                timer[15:8]<=nesprgdin;
            else if(timerEN)
                timer[15:8]<=timer_dec[15:8];

            if(nesprg_we & prgain[15:13]==3'b101 & regA==13 & ~nesprgdin[0]) // ack IRQ when IRQEN is cleared
                tripped<=0;
            else if(timerEN && timer_dec[16]) // trigger IRQ when timer wraps
                tripped<=1;
        end
    end
    assign irq=tripped & timerIRQEN;

//mirroring (00=V, 01=H, 10=1scA, 11=1scB)
    wire vertical= mirror==0;
    wire horizontal= mirror==1;
    wire oneB= mirror==3;
    assign ramchraout[10]=!chrain[13] ? chrbank[10] : (oneB | (vertical & chrain[10]) | (horizontal & chrain[11]));
    assign ramchraout[11]=chrbank[11];
    assign ciram_ce=chrain[13];

//rom size mask
    assign ramprgaout[18:13]=prgbank[18:13] & cfg_prgmask;
    assign ramchraout[18:12]=chrbank[18:12] & cfg_chrmask;

//bus control
    assign chrram_we=neschr_wr & !chrain[13] & cfg_chrram;
    assign chrram_oe=neschr_rd & !chrain[13];

    assign neschr_oe=0;
    assign neschrdout=8'bx;

    assign wram_oe=m2_n & ~nesprg_we & prgain[15:13]=='b011 & sramEN==3;
    assign wram_we=m2_n &  nesprg_we & prgain[15:13]=='b011 & sramEN==3;

    assign prgram_we=0;
    assign prgram_oe=~cfg_boot & m2_n & ~nesprg_we & (prgain[15] | (~sramEN[0] & prgain[15:13]=='b011));

    wire config_rd;
    gamegenie gg(m2, reset, nesprg_we, prgain, nesprgdin, ramprgdin, nesprgdout, config_rd);
    assign nesprg_oe=wram_oe | prgram_oe | config_rd;

//audio
    wire [6:0] fme7_out;
    FME7_sound snd0(m2, reset, nesprg_we, prgain, nesprgdin, fme7_out);
    pdm #(7) pdm_mod(clk20, fme7_out, exp6);

endmodule

module FME7_sound(
    input m2,
    input reset,
    input wr,
    input [15:0] ain,
    input [7:0] din,
    output [6:0] out
);
    reg [3:0] regC;
    reg [11:0] freq0,freq1,freq2;
    reg [2:0] en;
    reg [3:0] vol0,vol1,vol2;
    reg [11:0] count0,count1,count2;
    reg [4:0] duty0,duty1,duty2;
    always@(posedge m2, posedge reset) begin
        if(reset) begin
            en<=0;
        end else begin
            if(wr) begin
                if(ain[15:13]==3'b110)  //C000
                    regC<=din;
                if(ain[15:13]==3'b111)  //E000
                case(regC)
                    0:freq0[7:0]<=din;
                    1:freq0[11:8]<=din;
                    2:freq1[7:0]<=din;
                    3:freq1[11:8]<=din;
                    4:freq2[7:0]<=din;
                    5:freq2[11:8]<=din;
                    7:en<=din;
                    8:vol0<=din;
                    9:vol1<=din;
                    10:vol2<=din;
                endcase
            end
            if(count0==freq0) begin
                count0<=0;
                duty0<=duty0+1;
            end else
                count0<=count0+1;

            if(count1==freq1) begin
                count1<=0;
                duty1<=duty1+1;
            end else
                count1<=count1+1;
            if(count2==freq2) begin
                count2<=0;
                duty2<=duty2+1;
            end else
                count2<=count2+1;
        end
    end
    
    wire [3:0] ch0={4{~en[0] & duty0[4]}} & vol0;
    wire [3:0] ch1={4{~en[1] & duty1[4]}} & vol1;
    wire [3:0] ch2={4{~en[2] & duty2[4]}} & vol2;
    assign out=ch0+ch1+ch2;
    
endmodule
