`default_nettype none

// change regsel below to set VRC4 variant
module MAPVRC4(     //signal descriptions in powerpak.v
    input m2,
    input m2_n,
    input clk20,

    input reset,
    input nesprg_we,
    output nesprg_oe,
    input neschr_rd,
    input neschr_wr,
    input [15:0] prgain,
    input [13:0] chrain,
    input [7:0] nesprgdin,
    input [7:0] ramprgdin,
    output [7:0] nesprgdout,

    output [7:0] neschrdout,
    output neschr_oe,

    output chrram_we,
    output chrram_oe,
    output wram_oe,
    output wram_we,
    output prgram_we,
    output prgram_oe,
    output [18:10] ramchraout,
    output [18:13] ramprgaout,
    output irq,
    output ciram_ce,
    output exp6,
    
    input cfg_boot,
    input [18:12] cfg_chrmask,
    input [18:13] cfg_prgmask,
    input cfg_vertical,
    input cfg_fourscreen,
    input cfg_chrram
);
    //wire [5:0] regsel={prgain[15:12], prgain[2]|prgain[7], prgain[1]|prgain[6]};      //MAP15: VRC4a,c (21)
    //wire [5:0] regsel={prgain[15:12], prgain[3]|prgain[1], prgain[2]|prgain[0]};    //MAP17: VRC4e,VRC2b (23)
    wire [5:0] regsel={prgain[15:12], prgain[0]|prgain[2], prgain[1]|prgain[3]};    //MAP19: VRC4b,d (25)
    
    reg swap;
    reg [4:0] prg0,prg1;
    reg [7:0] chr0,chr1,chr2,chr3,chr4,chr5,chr6,chr7;
    reg [1:0] mirror;
    reg [7:0] irqlatch;
    reg irqM,irqE,irqA;
    always@(posedge m2) begin
        if(nesprg_we & prgain[15]) begin
            case(regsel[4:0])
                0,1,2,3:    prg0<=nesprgdin[4:0]; //8xxx
                
                4,5:        mirror<=nesprgdin[1:0]; //9xxx
                6,7:        swap<=nesprgdin[1];
                
                8,9,10,11:  prg1<=nesprgdin[4:0]; //Axxx

                12:     chr0[3:0]<=nesprgdin[3:0]; //Bxxx
                13:     chr0[7:4]<=nesprgdin[3:0];
                14:     chr1[3:0]<=nesprgdin[3:0];
                15:     chr1[7:4]<=nesprgdin[3:0];
                
                16:     chr2[3:0]<=nesprgdin[3:0]; //Cxxx
                17:     chr2[7:4]<=nesprgdin[3:0];
                18:     chr3[3:0]<=nesprgdin[3:0];
                19:     chr3[7:4]<=nesprgdin[3:0];
                
                20:     chr4[3:0]<=nesprgdin[3:0]; //Dxxx
                21:     chr4[7:4]<=nesprgdin[3:0];
                22:     chr5[3:0]<=nesprgdin[3:0];
                23:     chr5[7:4]<=nesprgdin[3:0];
                
                24:     chr6[3:0]<=nesprgdin[3:0]; //Exxx
                25:     chr6[7:4]<=nesprgdin[3:0];
                26:     chr7[3:0]<=nesprgdin[3:0];
                27:     chr7[7:4]<=nesprgdin[3:0];
                
                28:     irqlatch[3:0]<=nesprgdin[3:0]; //Fxxx
                29:     irqlatch[7:4]<=nesprgdin[3:0];
                30:     {irqM,irqA}<={nesprgdin[2],nesprgdin[0]};
            endcase
        end
    end

//bankswitch
    reg [17:13] prgbank;
    reg [17:10] chrbank;
    always@* begin
        case({swap,prgain[14],prgain[13]})
            0,6:prgbank<=prg0;
            1,5:prgbank<=prg1;
            2,4:prgbank<='b11110;
            3,7:prgbank<='b11111;
        endcase
        case(chrain[12:10])
            0:chrbank=chr0;
            1:chrbank=chr1;
            2:chrbank=chr2;
            3:chrbank=chr3;
            4:chrbank=chr4;
            5:chrbank=chr5;
            6:chrbank=chr6;
            7:chrbank=chr7;
        endcase
    end

//IRQ
    wire ack=regsel=='b111111;  //irq reg addrs
    wire ctl=regsel=='b111110;

    reg [7:0] irqcnt;
    reg timeout;
    reg [6:0] scalar;
    reg [1:0] line;
    wire irqclk=irqM|(scalar==0);
    wire setE=nesprg_we & ctl & nesprgdin[1];
    always@(posedge m2, posedge setE) begin
        if(setE) begin
            scalar<=113;
            line<=0;
            irqcnt<=irqlatch;
        end else if(irqE) begin
            if(scalar!=0)
                scalar<=scalar-1;
            else begin
                scalar<=(~line[1])?113:112;
                line<=line[1]?0:line+1;
            end
            if(irqclk) begin
                if(irqcnt==255)     irqcnt<=irqlatch;
                else            irqcnt<=irqcnt+1;
            end
        end
    end

    always@(posedge m2, posedge reset) begin
        if(reset) begin
            irqE<=0;
            timeout<=0;
        end else begin
            if(nesprg_we & (ack|ctl))
                timeout<=0;
            else if(irqclk & irqcnt==255)
                timeout<=1;

            if(nesprg_we & ctl) //control write
                irqE<=nesprgdin[1];
            else if(nesprg_we & ack) //irq acknowledge
                irqE<=irqA;
        end
    end

    assign irq=timeout & irqE;

//mirroring
    assign ramchraout[10]=!chrain[13] ? chrbank[10] : ((mirror==0 & chrain[10]) | (mirror==1 & chrain[11]) | (mirror==3));
    assign ramchraout[11]=chrbank[11];
    assign ciram_ce=chrain[13];

//rom size mask
    assign ramprgaout[18:13]={1'b0, prgbank[17:13] & cfg_prgmask[17:13]} & {4'b1111, {2{prgain[15]}}};
    assign ramchraout[18:12]={1'b0, chrbank[17:12] & cfg_chrmask[17:12]};

//ram control
    assign chrram_we=neschr_wr & !chrain[13] & cfg_chrram;
    assign chrram_oe=neschr_rd & !chrain[13];

    assign neschr_oe=0;
    assign neschrdout=8'bx;

    assign wram_oe=m2_n & ~nesprg_we & prgain[15:13]=='b011;
    assign wram_we=m2_n &  nesprg_we & prgain[15:13]=='b011;

    assign prgram_we=0;
    assign prgram_oe=~cfg_boot & m2_n & ~nesprg_we & prgain[15];

    wire config_rd;
    gamegenie gg(m2, reset, nesprg_we, prgain, nesprgdin, ramprgdin, nesprgdout, config_rd);
    assign nesprg_oe=wram_oe | prgram_oe | config_rd;

//sound
    assign exp6=1'bz;

endmodule
